/***********************************************************
 * $Id: BeanInvocationHandler.java 198 2010-09-06 12:05:18Z remoting $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 29.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.beans;

import java.io.InvalidClassException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.rmi.server.UID;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.CallbackInvocationContext;
import org.clazzes.remoting.Initializable;
import org.clazzes.remoting.InvocationContext;
import org.clazzes.remoting.InvocationHandler;
import org.clazzes.remoting.RemoteInvocation;
import org.clazzes.remoting.RemoteInvocationFailureException;

/**
 * This ServerInvocationHandler delegates remote calls to a bean by
 * intelligently discovering the service interface.
 * 
 * @author wglas
 */
public class BeanInvocationHandler implements InvocationHandler, Initializable {

	private static final Log log = LogFactory.getLog(BeanInvocationHandler.class);

	private Object target;
    private java.lang.reflect.InvocationHandler targetInvocationHandler;
	private Class<?> serviceInterface;
	private ClassLoader callbackProxyClassLoader;

	private static class MethodInfo {
		
		public final Method method;
		public final int [] callbackArgs;
		
		public MethodInfo(Method method, int[] callbackArgs) {
			super();
			this.method = method;
			this.callbackArgs = callbackArgs;
		}
	}
	
	private Map<String,MethodInfo> serviceMethods;
	private List<Class<?>> callbackInterfaces;
		
	private ReturnCallbackHandler returnCallbackHandler;
	private BeanCallAuthManager beanCallAuthManager;
	
	private class CallbackInvocationHandler implements java.lang.reflect.InvocationHandler {

		final CallbackInvocationContext context;
		final UID callbackId;
		
		public CallbackInvocationHandler(CallbackInvocationContext context, UID callbackId) {
			super();
			this.context = context;
			this.callbackId = callbackId;
		}

		public synchronized Object invoke(Object proxy, Method method, Object[] args)
				throws Throwable {
				
			if (log.isDebugEnabled())
				log.debug("Issuing callback with Id ["+this.callbackId+"] in thread ["+Thread.currentThread().getName()+"].");

			try {
				Object res = this.context.callBack(this.callbackId,new RemoteInvocation(method.getName(),args),null);
				
				return res;
			} catch (InvocationTargetException ite) {
			
				// combine the foreign and the local stack trace.
				Throwable e = ite.getCause();
				
				StackTraceElement[] foreignStack = e.getStackTrace();
				e.fillInStackTrace();
				StackTraceElement[] myStack = e.getStackTrace();
				StackTraceElement[] combinedStack = new StackTraceElement[foreignStack.length+myStack.length];
				System.arraycopy(foreignStack,0,combinedStack,0,foreignStack.length);
				System.arraycopy(myStack,0,combinedStack,foreignStack.length,myStack.length);
				e.setStackTrace(combinedStack);
				throw e;			
			}
		}

	}

	/* (non-Javadoc)
     * @see org.clazzes.remoting.Initializable#isInitialized()
     */
    public boolean isInitialized() {
        
        return this.serviceMethods != null;
    }

    /* (non-Javadoc)
	 * @see org.clazzes.remoting.Initializable#initialize()
	 */
	public void initialize() throws Exception {
		
		if (this.serviceInterface == null)
		{
			Class<?>[] ifaces = this.target.getClass().getInterfaces();
			
			if (ifaces.length <= 0)
				throw new InvalidClassException("Target of class ["+this.target.getClass()+"] does not implement a single interface.");
			
			this.serviceInterface = ifaces[0];
		}
		
		if (log.isDebugEnabled())
			log.debug("Using service interface ["+this.serviceInterface+"].");
		
		Method [] methods = this.serviceInterface.getMethods();
		
		this.serviceMethods = new HashMap<String, MethodInfo>();
		
		for (Method method : methods) {
			
			MethodInfo om = this.serviceMethods.get(method.getName());
			
			if (om != null)
				throw new InvalidClassException("Method ["+om.method+"] is duplicated in interface ["+this.serviceInterface.getName()+"] (duplicate is ["+method+"]).");
			
			int callbackArgs[] = null;
			int nCallBackArgs = 0;

			if (this.callbackInterfaces != null) {
			
				for (int iarg = 0; iarg < method.getParameterTypes().length; ++iarg){
					
					Class<?> arg = method.getParameterTypes()[iarg];
					
					for (Class<?> iface : this.callbackInterfaces) {
					
						if (iface.isAssignableFrom(arg))
						{
							if (callbackArgs == null)
								callbackArgs = new int[method.getParameterTypes().length];
							
							if (log.isDebugEnabled()) 
								log.debug("Argument ["+iarg+"] of method ["+method+"] will be treated as callback with interface ["+iface.getName()+"].");
							
							callbackArgs[nCallBackArgs] = iarg;
							++nCallBackArgs;
						}
					}
				}	
			}
			
			if (callbackArgs != null) {
				
				int x[] = new int[nCallBackArgs];
				
				System.arraycopy(callbackArgs,0,x,0,nCallBackArgs);
				callbackArgs=x;
			}
			
			if (log.isDebugEnabled())
				log.debug("Adding method ["+method +"] to the list of service methods.");
			
			this.serviceMethods.put(method.getName(),new MethodInfo(method,callbackArgs));
		}
	}

	/* (non-Javadoc)
	 * @see org.clazzes.remoting.InvocationHandler#invoke(org.clazzes.remoting.InvocationContext, java.lang.Object, java.util.Map)
	 */
	public Object invoke(InvocationContext ctx, Object arg,
			Map<String, Object> metadata) throws Throwable {
	
		if (arg instanceof RemoteInvocation) {
			RemoteInvocation ri = (RemoteInvocation) arg;
			
			MethodInfo m = this.serviceMethods.get(ri.getMethodName());
			
			if (m == null)
				throw new RemoteInvocationFailureException("The service interface ["+this.serviceInterface.getName()+"] contains no method named ["+ri.getMethodName()+"].");

			if (log.isDebugEnabled())
				log.debug("Calling method ["+ri.getMethodName()+"] on interface ["+this.serviceInterface.getSimpleName()+"]...");
			
			try {
				
				Object[] args;
				
				if (m.callbackArgs == null) {
					
					args = ri.getArguments();
				} else {
					
					Object[] params = ri.getArguments();
					args = new Object[params.length];
					
					int ilast = 0;
					
					int nCallbacks = 0;
					
					for (int iarg : m.callbackArgs) {
						
						if (params[iarg] != null) {
						
							if (iarg > ilast)
							System.arraycopy(params,ilast,args,ilast,iarg-ilast);
						
							if (log.isDebugEnabled())
								log.debug("Introducing callback proxy no . ["+nCallbacks+"] for argument ["+iarg+"] with UID ["+params[iarg]+"]...");

							ClassLoader cl = this.callbackProxyClassLoader;
							
							if (cl == null)
							    cl = m.method.getParameterTypes()[iarg].getClassLoader();
							
							args[iarg] = Proxy.newProxyInstance(cl,
									new Class[]{m.method.getParameterTypes()[iarg]},
									new CallbackInvocationHandler((CallbackInvocationContext)ctx,(UID)params[iarg]));
						
							ilast = iarg+1;
							++nCallbacks;
						}
					}
					if (params.length > ilast)
						System.arraycopy(params,ilast,args,ilast,params.length-ilast);
				}
				
				if (this.beanCallAuthManager != null)
				    this.beanCallAuthManager.authenticate(ri,m.method,ctx.getClientUID(),metadata);
				
				Object ret;
				
				if (this.targetInvocationHandler == null)
				    ret = m.method.invoke(this.target,args);
				else
				    ret = this.targetInvocationHandler.invoke(this.target,m.method,args);
			
				if (log.isDebugEnabled())
					log.debug("Sucessfully called method ["+ri.getMethodName()+"] on interface ["+this.serviceInterface.getSimpleName()+"].");
				
				if (ret != null && this.returnCallbackHandler != null && this.returnCallbackHandler.isCallbackInterface(m.method.getReturnType()))
					ret = this.returnCallbackHandler.registerCallback(ctx.getClientUID(),m.method.getReturnType(),ret);
				
				return ret;
			}
			catch (InvocationTargetException e) {
				
				if (log.isDebugEnabled())
					log.debug("Caught exception in method ["+ri.getMethodName()+"] of interface ["+this.serviceInterface.getSimpleName()+"].",e.getCause());
				throw e.getCause();
			}
		}
		
		throw new RemoteInvocationFailureException("The invocation parameter ["+arg +"] is not of type org.jboss.remoting.RemoteInvocation");
	}

	/**
	 * @return The target bean on which to invoke remote invocations.
	 */
	public Object getTarget() {
		return this.target;
	}

	/**
	 * @param target The target bean to set.
	 */
	public void setTarget(Object target) {
		this.target = target;
		
		if (Proxy.isProxyClass(target.getClass())) {
		    this.targetInvocationHandler = Proxy.getInvocationHandler(target);
		}
		else {
		    this.targetInvocationHandler = null;
		}
	}

	/**
	 * @return the serviceInterface. If the service interface is not set, the first interface of
	 *         the target will be used.
	 */
	public Class<?> getServiceInterface() {
		return this.serviceInterface;
	}

	/**
	 * @param serviceInterface the serviceInterface to set.
	 * @throws InvalidClassException If the parameter is a class rather than an interface.
	 */
	public void setServiceInterface(Class<?> serviceInterface) throws InvalidClassException {
		
		if (!serviceInterface.isInterface())
			throw new InvalidClassException("Class ["+serviceInterface.getName()+"] is not an interface.");
		
		this.serviceInterface = serviceInterface;
	}

	/**
	 * @return the callbackInterfaces
	 */
	public List<Class<?>> getCallbackInterfaces() {
		return this.callbackInterfaces;
	}

	/**
	 * @param callbackInterfaces the callbackInterfaces to set
	 */
	public void setCallbackInterfaces(List<Class<?>> callbackInterfaces) {
		this.callbackInterfaces = callbackInterfaces;
	}

	/**
     * @return the class loader to pass to
     *         {@link Proxy#newProxyInstance(ClassLoader, Class[], InvocationHandler)},
     *         when instantiating callback proxies.
     *         If <code>null</code>, <code>callbackInterface.class.getClassLoader()</code> is used
     *         for defining callback proxy classes.
     */
    public ClassLoader getCallbackProxyClassLoader() {
        return this.callbackProxyClassLoader;
    }

    /**
     * @param callbackProxyClassLoader the class loader to pass to
     *         {@link Proxy#newProxyInstance(ClassLoader, Class[], InvocationHandler)},
     *         when instantiating callback proxies.
     *         If <code>null</code>, <code>callbackInterface.class.getClassLoader()</code> is used
     *         for defining callback proxy classes.
     */
    public void setCallbackProxyClassLoader(ClassLoader callbackProxyClassLoader) {
        this.callbackProxyClassLoader = callbackProxyClassLoader;
    }

    /**
	 * @return the returnCallbackHandler
	 */
	public ReturnCallbackHandler getReturnCallbackHandler() {
		return this.returnCallbackHandler;
	}

	/**
	 * @param returnCallbackHandler the returnCallbackHandler to set
	 */
	public void setReturnCallbackHandler(ReturnCallbackHandler returnCallbackHandler) {
		this.returnCallbackHandler = returnCallbackHandler;
	}

    /**
     * @return the beanCallAuthManager
     */
    public BeanCallAuthManager getBeanCallAuthManager() {
        return this.beanCallAuthManager;
    }

    /**
     * @param beanCallAuthManager the beanCallAuthManager to set
     */
    public void setBeanCallAuthManager(BeanCallAuthManager beanCallAuthManager) {
        this.beanCallAuthManager = beanCallAuthManager;
    }
}
