/***********************************************************
 * $Id: ClientHolder.java 195 2010-09-06 06:08:19Z remoting $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 29.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.beans;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.concurrent.TimeoutException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.client.Client;
import org.clazzes.remoting.client.ClientManager;
import org.clazzes.util.aop.ThreadLocalManager;

/**
 * This class provides static helper functions for thread-bound session contexts, which
 * may be used by session interceptors or servlet filters.
 */
public abstract class ClientHolder {

    private static final Log log = LogFactory.getLog(ClientHolder.class);
    
    /**
     * Get a thread-bound client for a client manager, which has previously been opened with
     * {@link #openContext(ClientManager, String, SessionAuthenticator)}.
     * 
     * @param clientManager The client manager for which to open the session.
     * @return The client holder attached to the current thread or <code>null</code>,
     *         if no context is bound to the current thread. 
     */
    public static Client getContext(ClientManager clientManager) {
        
        return ThreadLocalManager.getBoundResource(clientManager.getUniqueID());
    }
    
    private static void openSession(Client client,
            String sessionContextSubSystem,
            SessionAuthenticator sessionAuthenticator) throws InvocationTargetException, IOException {

        if (sessionAuthenticator == null) {
            
            client.invoke(sessionContextSubSystem,SessionCmd.OPEN_SESSION,null);
        }
        else {
            
            client.invoke(sessionContextSubSystem,SessionCmd.OPEN_SESSION,
                    sessionAuthenticator.createCredentials(client.getClientUID()));
        }

    }
    
    private static void closeSession(Client client, String sessionContextSubSystem) throws InvocationTargetException, IOException {
        client.invoke(sessionContextSubSystem,SessionCmd.CLOSE_SESSION,null);
    }
  
    /**
     * Open a thread-bound session context for the given client manager.
     * 
     * @param clientManager The client manager for which to open the session.
     * @param sessionContextSubSystem The subsystem name of the remote session context subsystem.
     * @param sessionAuthenticator An optional session authenticator, which may be null.
     * @return The new client holder.
     * @throws TimeoutException If no more clients are available.
     * @throws IOException If an unrecoverable transport error occurred.
     * @throws InvocationTargetException If the remote system threw an exception upon opening the sessoin context.
     */
    public static Client openContext(ClientManager clientManager, 
            String sessionContextSubSystem,
            SessionAuthenticator sessionAuthenticator) throws TimeoutException, IOException, InvocationTargetException {
        
        Client client = ThreadLocalManager.getBoundResource(clientManager.getUniqueID());
        
        if (client != null) {
            
            log.warn("Closing left-over session for current thread ["+Thread.currentThread().getName()+"] before trying to open session.");
            closeContextInner(client,clientManager,sessionContextSubSystem);
        }
        
        client = clientManager.activateClient();
        
        if (log.isDebugEnabled())
            log.debug("Binding client session context to current thread ["+Thread.currentThread().getName()+"].");

        try {
            
            if (log.isDebugEnabled())
                log.debug("Opening remote session for current thread ["+Thread.currentThread().getName()+"].");         

            try {
                openSession(client, sessionContextSubSystem, sessionAuthenticator);
            } catch (IOException e) {
                log.warn("Transport error while opening session, trying to reconnect.",e);
                client.disconnect();
                client.connect();
                openSession(client, sessionContextSubSystem, sessionAuthenticator);
                log.info("Successfully reconnected client.");
            }
            
        }
        catch (Throwable e) {
            client.disconnect();
            clientManager.passivateClient(client);
        }
        
        ThreadLocalManager.bindResource(clientManager.getUniqueID(),client);
        return client;
    }
    
    /**
     * Close a thread-bound remote session context for the given client manager.
     * 
     * @param clientManager The client manager for which to close the remote session.
     * @param sessionContextSubSystem The subsystem name of the remote session context subsystem.
     */
    public static void closeContext(ClientManager clientManager, 
            String sessionContextSubSystem) {
        
        Client client = ThreadLocalManager.unbindResource(clientManager.getUniqueID());
        
        if (client == null) {
            
            log.warn("No session context for current thread ["+Thread.currentThread().getName()+"] when trying to close session.");
            return;
        }
        
        closeContextInner(client,clientManager,sessionContextSubSystem);
    }
    
    private static void closeContextInner(Client client,
            ClientManager clientManager, 
            String sessionContextSubSystem) {
            
        if (client.isConnected()) {
            try {
                if (log.isDebugEnabled())
                    log.debug("Closing remote session of current thread ["+Thread.currentThread().getName()+"].");           
                
                closeSession(client,sessionContextSubSystem);
            }
            catch (Throwable e) {
                log.warn("Error closing remote seession of current thread ["+Thread.currentThread().getName()+"].",e);
            }
        }
                
        clientManager.passivateClient(client);
        
        if (log.isDebugEnabled())
            log.debug("Unbinding client session context Session from current thread ["+Thread.currentThread().getName()+"].");          
    }
}
