/***********************************************************
 * $Id: ReturnCallbackImporter.java 195 2010-09-06 06:08:19Z remoting $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 29.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.beans;

import java.io.IOException;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.RemoteInvocation;
import org.clazzes.remoting.client.Client;
import org.clazzes.remoting.client.ClientManager;

/**
 * Creates import proxies exported by {@link ReturnCallbackInvocationHandler} via remoting.
 * 
 * @author wglas
 */
public class ReturnCallbackImporter implements ReturnCallbackProxyFactory {

	private static final Log log = LogFactory.getLog(ReturnCallbackImporter.class);
		
	private ClientManager clientManager;
	private Set<Class<?>> callbackInterfaces;
	private ClassLoader callbackProxyClassLoader;

	private String subSystem;
	
	private class ReturnInvocationHandler implements InvocationHandler {

		private final Map<String,Object> payload;
		private final Object proxyGUID;
		
		public ReturnInvocationHandler(Object proxyGUID) {
			
			this.proxyGUID = proxyGUID;
			this.payload = new HashMap<String, Object>(3); 
				
			this.payload.put(ReturnCallbackHandler.RETURN_PROXY_GUID_PROPERTY,proxyGUID);
		}

		/* (non-Javadoc)
		 * @see java.lang.reflect.InvocationHandler#invoke(java.lang.Object, java.lang.reflect.Method, java.lang.Object[])
		 */
		public Object invoke(Object proxy, Method method, Object[] args)
				throws Throwable {
			
            Client client = ClientHolder.getContext(ReturnCallbackImporter.this.clientManager);
            
            boolean threadBoundClient;
            
            if (client == null) {
                threadBoundClient = false;
                client = ReturnCallbackImporter.this.clientManager.activateClient();
            }
            else {
                threadBoundClient = true;
            }
			
			try {

				Object res;
				
				try {
					
					if (log.isDebugEnabled())
						log.debug("Calling method ["+method.getName()+"] on subsystem ["+ReturnCallbackImporter.this.subSystem+"], GUID ["+this.proxyGUID+"]...");
					
					res = client.invoke(ReturnCallbackImporter.this.subSystem,new RemoteInvocation(method.getName(),args),this.payload);
										
					if (log.isDebugEnabled())
						log.debug("Successfully called method ["+method.getName()+"] on subsystem ["+ReturnCallbackImporter.this.subSystem+"], GUID ["+this.proxyGUID+"].");

				} catch (IOException t) {
					
				    log.error("Disconnecting client upon transport error.",t);
				    client.disconnect();
					
					throw t;
				}
					
				catch (InvocationTargetException ite) {
					
					// combine the foreign and the local stack trace.
					Throwable e = ite.getCause();
					
					StackTraceElement[] foreignStack = e.getStackTrace();
					e.fillInStackTrace();
					StackTraceElement[] myStack = e.getStackTrace();
					StackTraceElement[] combinedStack = new StackTraceElement[foreignStack.length+myStack.length];
					System.arraycopy(foreignStack,0,combinedStack,0,foreignStack.length);
					System.arraycopy(myStack,0,combinedStack,foreignStack.length,myStack.length);
					e.setStackTrace(combinedStack);
					throw e;			
				}
				
				return res;
				
			} finally {
				if (!threadBoundClient)
					ReturnCallbackImporter.this.clientManager.passivateClient(client);
			}
		}
	}
	
	/* (non-Javadoc)
	 * @see at.cdes.remoting.client.ReturnCallbackProxyFactory#cerateProxy(java.lang.Class, java.lang.Object)
	 */
	public Object createProxy(Class<?> iface, Object ret) throws Exception {
		
		if (this.callbackInterfaces.contains(iface)) {

			if (log.isDebugEnabled())
				log.debug("Creating proxy for interface ["+iface.getName()+"], GUID ["+ret+"].");
				
			ClassLoader cl = this.callbackProxyClassLoader;
			
			if (cl == null)
			    cl = iface.getClassLoader();
			    
			return
			Proxy.newProxyInstance(cl,
					new Class<?>[] {iface},new ReturnInvocationHandler(ret));
		}
		else
			return ret;
	}

	/**
	 * @return the clientManager
	 */
	public ClientManager getClientManager() {
		return this.clientManager;
	}

	/**
	 * @param clientManager the clientManager to set
	 */
	public void setClientManager(ClientManager clientManager) {
		this.clientManager = clientManager;
	}

	/**
	 * @return the jboss remoting subsystem.
	 * @see Client#setSubsystem(String)
	 */
	public String getSubSystem() {
		return this.subSystem;
	}

	/**
	 * @param subSystem the subSystem to set
	 */
	public void setSubSystem(String subSystem) {
		this.subSystem = subSystem;
	}

	/**
	 * @return The set of callback interfaces, which are proxied by this instance.
	 * 
	 * @see ReturnCallbackInvocationHandler#setCallbackInterfaces(java.util.List)
	 */
	public Set<Class<?>> getCallbackInterfaces() {
		return this.callbackInterfaces;
	}

	/**
	 * @param serviceInterfaces The set of callback interfaces, which are proxied by this instance.
	 * 
	 * @see ReturnCallbackInvocationHandler#setCallbackInterfaces(java.util.List)
	 */
	public void setCallbackInterfaces(Set<Class<?>> serviceInterfaces) {
		this.callbackInterfaces = serviceInterfaces;
	}

    /**
     * @return the class loader to pass to
     *         {@link Proxy#newProxyInstance(ClassLoader, Class[], InvocationHandler)},
     *         when instantiating callback proxies.
     *         If <code>null</code>, <code>callbackInterface.class.getClassLoader()</code> is used
     *         for defining callback proxy classes.
     */
    public ClassLoader getCallbackProxyClassLoader() {
        return this.callbackProxyClassLoader;
    }

    /**
     * @param callbackProxyClassLoader the class loader to pass to
     *         {@link Proxy#newProxyInstance(ClassLoader, Class[], InvocationHandler)},
     *         when instantiating callback proxies.
     *         If <code>null</code>, <code>callbackInterface.class.getClassLoader()</code> is used
     *         for defining callback proxy classes.
     */
    public void setCallbackProxyClassLoader(ClassLoader callbackProxyClassLoader) {
        this.callbackProxyClassLoader = callbackProxyClassLoader;
    }
}
