/***********************************************************
 * $Id: ObjectMarshaler.java 144 2009-10-12 07:47:02Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 23.08.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.marshal.impl;

import java.io.IOException;
import java.lang.reflect.Method;
import java.util.Date;
import java.util.GregorianCalendar;

import org.clazzes.remoting.marshal.ObjectResolver;

/**
 * This interface is used for reading an object from a compact marshaler
 * after reading a class header.
 * 
 * The are several classes like {@link Class}, {@link Date}, {@link GregorianCalendar}
 * which may not be handled by simply instantiating an instance and reading the
 * non-transient fields. Such classes have a specialized implementation of ObjectMarshaler
 * apart from the default implementation {@link ClassFields}.
 * 
 * An implementation might as well implement the {@link ObjectResolver} interface, which
 * may be used to call <code>this.readResolve()</code> and <code>this.writeReplace()</code>
 * on the deserialized instance.
 */
interface ObjectMarshaler {

    /**
     * Instantiate a new instance of the given class info and read the fields.
     * 
     * The implementation of this method is responsible for calling
     * {@link CompactMarshaler#cacheReadPremature(long, Object)} right after instantiating
     * the new object instance and before reading the fields.
     *  
     * This is necessary in order to handle references to the deserialized object
     * in the object tree.
     * 
     * @param context The marshaler at whom's behalf we are acting.
     * @param ref The object reference from the object header.
     * @param ci The class information from the header.
     * @return A deserialized object.
     * @throws IllegalArgumentException
     * @throws IllegalAccessException
     * @throws IOException
     * @throws ClassNotFoundException
     */
    public Object readObject(CompactMarshaler context, long ref, ClassInfo ci) throws IllegalArgumentException, IllegalAccessException, IOException, ClassNotFoundException;
    
    /**
     * Write the fields of the given object to the 
     * @param context The marshaler at whom's behalf we are acting.
     * @param obj The object to be written.
     * @throws IllegalArgumentException
     * @throws IOException
     * @throws IllegalAccessException
     * @throws ClassNotFoundException
     */
    public void writeObject(CompactMarshaler context, Object obj) throws IllegalArgumentException, IOException, IllegalAccessException, ClassNotFoundException;

    /**
     * @return The <code>writeReplace()</code> method of the represented class or null,
     *         if this class has not write replacement.
     */
    public Method getWriteReplaceMethod();

    /**
     * @return The <code>readResolve()</code> method of the represented class or null,
     *         if this class has not read replacement.
     */
    public Method getReadResolveMethod();
}
